/*
*	login.c
*
*	History:
*	1993/8	Rebuilded by Kenar
*	1995/8	Bugfix in sort routine by Kenar
*/
#include "tip.h"
#include <sys/time.h>

extern char *skip(char*);
extern void rmEol(char*);
extern int sepstr(char* str,char** argp,char* sep);

int scnextln; // This value is initialized in value.c

static int vmode; // visible mode controled with "+v" & "-v" in host.tipp
		  //		 -- Kenar

typedef struct
{	int lineNo;
	char *cmdLn;
	int argc;
	char **argv;
}	ScriptLineType;

#define MAXLINE 512
#define LINESIZ 128
#define ERR (-1)

static char* execScript();
static char *dosend();
sigfunc_t naptimeout();
static char *scriptErrMsg();
static char *dowait();
static int getch();
static int waitStrs();
static void sort();
static int dogoto();
static char _got[80];

char *login(fp)
FILE *fp;
{	char buf[LINESIZ+2];
	int n,k,i,line;
	ScriptLineType sLine[MAXLINE];
	char *tmpArg[LINESIZ/2];
	char *p;

	// Replaced by Kenar
	// Read script file and store the data to memory
	line = scnextln; // line number for error message
	n = 0;    // n is the effective script line number
	while((fgets(buf, sizeof(buf), fp) != NULL) 
		&& (strncmp(buf,"%end",4) != 0))
	{
		line++;
		rmEol(buf); // remove EOL at end
		p = skip(buf);
		if(!*p) continue;
		if (*p == '#') continue; // comment line
		
		/* Don,t label conversion move here.
		   Because we have a case such like:
		   wait 5/error login:
		*/
		
		sLine[n].lineNo = line;
		sLine[n].cmdLn = makstr(p);
		if (strncmp(p,"echo",4) == 0)
		{	sLine[n].argc = 2;
			sLine[n].argv = (char **)calloc(3,sizeof(char*));
			sLine[n].cmdLn[4] = 0;
			sLine[n].argv[0] = sLine[n].cmdLn;
			sLine[n].argv[1] = sLine[n].cmdLn + 5;
			sLine[n].argv[2] = 0;
		}
		else
		{	sLine[n].argc = k = 
				sepstr(sLine[n].cmdLn,tmpArg," \t");
			sLine[n].argv = (char**)calloc(k + 1,sizeof(char*));
			for(i = 0; i < k ; i++)
				sLine[n].argv[i] = tmpArg[i];
			sLine[n].argv[k] = NULL;

			/* convert "LABEL:" to "label LABEL"  */
			p = sLine[n].argv[0];
			if(p[(k = strlen(p)-1)] == ':')
			{	p[k] = '\0';
				sLine[n].argv =
					(char**)calloc(3,sizeof(char*));
				sLine[n].argv[1] = p;
				sLine[n].argv[0] = "label";
				sLine[n].argv[2] = NULL;
				sLine[n].argc = 2;
			}
		}		
		n++;
	}
	p = execScript(sLine,n); // n is line count
	return p;
}

/* added by Kenar */
void echo(s)
char *s;
{	char *p,*s0;
	s0 = s;
	while((p = index(s,'@')) != NULL)
	{	if(p != s0 && *(p-1) == '\\') s = p + 1;
		else
		{	*p = 0;
			printf("%s%s",s,_got);
			*p++ = '@';
			s = p;
		}
	}
	puts(s);
}

static char* execScript(sLine,lineCount)
ScriptLineType *sLine;
int lineCount;
{	int n,line;
	int cnt;
	char *cmd,*cp,**q;
	ScriptLineType *p;

	for(n = 0; n < lineCount; n++)
	{	p = &sLine[n];
		cmd = p->argv[0];
		cnt = p->argc;
		line = p->lineNo;
		if(vflag){	int k;
			printf("%3d:",line);
			for(k = 0; p->argv[k]; k++) printf(" %s",p->argv[k]);
			putchar('\n');
		}
		if (equal(cmd, "done")) return(NULL);
		else if (equal(cmd, "fail")) return("login failed");
		else if(equal(cmd, "send"))
		{	if(cnt != 2) return(scriptErrMsg(line));
			if (cp = dosend(p->argv[1])) return(cp);
		}
		else if(equal(cmd, "wait"))
		{	q = p->argv;
			cp = dowait(++q);
					// label name is returned
			if(!cp) return(scriptErrMsg(line));
			if(!equal(cp,""))
			{	n = dogoto(cp,sLine,lineCount);
				if(n < 0)
				{	fprintf(stderr,
						"label %s not found\n",cp);
					return(scriptErrMsg(line));
				}
			}
		}
		else if(equal(cmd, "label")) continue;
		else if(equal(cmd, "goto"))
		{	if(cnt != 2) return(scriptErrMsg(line));
			n = dogoto(p->argv[1],sLine,lineCount);
			if(n < 0) return(scriptErrMsg(line));
		}
		else if(equal(cmd,"echo")) echo(p->argv[1]);
		else if(equal(cmd,"+v"))
		{	vmode = 1;
			continue;
		}
		else if(equal(cmd,"-v"))
		{	vmode = 0;
			continue;
		}
		else
			return(scriptErrMsg(line));
	}
	alarm(0);
	if (vflag) putc('\n', stderr);
	return(NULL);
}

static
char* scriptErrMsg(line) int line;
{	static char errmsg[80];
	sprintf(errmsg, "Error in %.80s at line %d", LS, line);
	return(errmsg);
}

static
char *dosend(cp)
char *cp;
{
	char c;

	if(vflag)
		fprintf(stderr, "send: %s: ", cp);
	while(*cp) {
		c = *cp++;
		if (c == '\\') {
			switch(*cp) {
			case 0:
				return("send format error");
			case 'b':
				c = '\b'; cp++; break;
			case 'd':
				sleep(1); cp++;
				continue;
			case 'f':
				c = '\f'; cp++; break;
			case 'n':
				c = '\n'; cp++; break;
			case 'r':
				c = '\r'; cp++; break;
			case 's':
				c = ' '; cp++; break;
			case 't':
				c = '\t'; cp++; break;
			case 'x':
				genbrk(); cp++;
				continue;
			case '?':
				c = 0177; cp++; break;
			default:
				c = octalnum(&cp);
			}
		}
		if (pwrite(&c, 1) != 1)
			return("write error to remote site");
		if (vflag)
			show(&c, 1);
		nap(20);
	}
	if (vflag)
		putc('\n', stderr);
	return(0);
	
}

nap(ms)
int ms;
{
	struct itimerval itv;
	signal(SIGALRM, naptimeout);
	itv.it_interval.tv_sec = 0;
	itv.it_interval.tv_usec = 0;
	itv.it_value.tv_sec = 0;
	itv.it_value.tv_usec = ms*1000;
	setitimer(ITIMER_REAL, &itv, (struct itimerval *)0);
	sigpause(0);
}

sigfunc_t naptimeout()
{
	/* return(0); */
}

jmp_buf jmpbuf;
sigfunc_t waittimeout();

static 
char *dowait(argv)	// returns label name
char **argv;
/*	dowait returns label name.
	if no label then returns NULL string; (no label is also legal syntax)
	if error then returns NULL */
{
	lett *strs[128],*labs[128];
	lett *label,*s;
	lett *control = argv[0];
	int secs = atoi(control);
	int j,n,m;
	
	if (label = strchr(control, '/'))
		label++;
	if (setjmp(jmpbuf)) {
		if(vflag) fprintf(stderr,
				"\nwait: timeout, going to %s\n", label);
		strcpy(_got,"TIME OUT");
		alarm(0);
		return(label);
	}
	signal(SIGALRM, waittimeout);
	alarm(secs);
	
	argv++;
	n = 0;
	while(*argv)
	{	m = n;
		while(*argv && **argv != '/')
		{	strs[n] = *argv;
			n++; argv++;
		}
		
		if(m == n) /* syntax err: no string to wait */
			return NULL;
		if(*argv) /* then *argv is label */
		{	s = *argv;
			s++;
		}
		else	/* no label */
			s = "";
		
		while(m < n)
		{	labs[m] = s;
			m++;
		}
		
		if(!*argv) break;

		argv++;
	}
	
	if(vflag)
	{	fprintf(stderr,"Waiting for %d strings ... ",n);
		for(j = 0; j < n; j++)
		{	fputs(strs[j],stderr);
			putc(' ',stderr);
		}
		putc('\n',stderr);
	}
	
	for(j = 0; j < n; j++) strfix(strs[j]);
	
	/*	------------ SORT ROUTINE -----------------	*/
	/* 	strs should be sorted				*/
	/*		before handed over to waitStrs		*/
	/*	STRAIGHT SELECTION
				*/
	/*	ref: Wirth,"Algorithm and Data Structure"	*/
	/*	external valiable: n,strs,labs	*/
	/*	coded by Kenar					*/
	{	int i,j,k;
		lett *x,*t;

		for(i = 0; i < n - 1; i++){
			k = i;
			for(j = i + 1; j < n; j++)
				if(strcmp(strs[j],strs[k]) < 0) k = j;
			/* swap i and k */
			x = strs[i]; strs[i] = strs[k]; strs[k] = x;
			t = labs[i]; labs[i] = labs[k]; labs[k] = t;
		}

	}

	n = waitStrs(n,strs);
	alarm(0);
	if(n == ERR) return NULL;
	if(vflag) fprintf(stderr,"\nRECEIVED %s\n",strs[n]);
	strcpy(_got,strs[n]);
	return(labs[n]);
}


static int _ungetcount;
static lett _ungetbuf[512];

/*
*	getch() read a char from modem
*/
static
int getch()
{	lett buf[1];
	if(_ungetcount > 0)
	{	_ungetcount--;
		return _ungetbuf[_ungetcount];
	}
	else if(pread(buf,1) < 1)
	{	fputs("Error in reading device\n",stderr);
		return ERR;
	}
	else
	{	if(vflag) show(buf,1);
		else if(vmode) putc(buf[0],stderr);
		return buf[0];
	}
}
	
static
void ungetch(s,n)
lett *s;
int n;
{	while(n > 0)
	{	n--;
		_ungetbuf[_ungetcount] = s[n];
		_ungetcount++;
	}
}
	
static
int waitStrs(n,strs)
int n;  // count of string
lett **strs; // string pointer
/*	coded by Kenar		*/
{	int j,k;

	int c,m;

	lett buf[512];
	int lenStr[128];
	
	if(n == 0) for(;;) c = getch(); // wait forever discarding data.

	for(j = 0; j < n; j++) lenStr[j] = strlen(strs[j]);
	

	/*	<<Search strings in a stream>>		 *		---- coded by Kenar ----
	 *
	 *	Target strings must be sorted.		 *
	 *	Be careful with the following example: 	 *		Stream : AAAAB
	 *		target : AAAB
	 *	This example show:
	 *		when the stream	data begin
	 *		matching with target, they
	 *		should be stocked in stack,
	 *		and when difference is found	 *		comparison should be started 	 *		from second character in the 	 *		stack.
	 *	In this codes the stack is named buf.	 */
	for(;;)
	{	m = 0; j = 0; k = n - 1;
		while(j < k)
		{	c = getch();
			if(c == ERR) break;
			while (j < n && c > strs[j][m]) j++;
			while (k >= 0 && c < strs[k][m]) k--;
			buf[m++] = c;
		}
		if ( j == k )
		{   while(m < lenStr[j])
			{	c = getch();
				if(c == ERR) break;
				buf[m] = c;
				if(strs[j][m] != c) break;
				m++;
			}
			if(m == lenStr[j]) return j; /* found */
			if(m > 1) ungetch(&buf[1],m - 1);
		}
		if(c == ERR) return ERR;
	}
}



strfix(cp)
char *cp;
{
	char c;
	char *np = cp;

	while(*cp) {
		c = *cp++;
		if (c == '\\') {
			switch(*cp) {
			case 'b':
				c = '\b'; cp++; break;
			case 'f':
				c = '\f'; cp++; break;
			case 'n':
				c = '\n'; cp++; break;
			case 'r':
				c = '\r'; cp++; break;
			case 's':
				c = ' '; cp++; break;
			case 't':
				c = '\t'; cp++; break;
			case '?':
				c = 0177; cp++; break;
			default:
				c = octalnum(&cp);
			}
		}
		*np++ = c; 
	}
	*np = 0;
}

#define isoctal(c) ((c) >= '0' && (c) <= '7')

octalnum(cpp)
char **cpp;
{
	int n = 0;
	char c = *((*cpp)++);

	if (isoctal(c)) {
		n = (c-'0');
		c = *((*cpp)++);
		if (isoctal(c)) {
			n = (n<<3)+(c-'0');
			c = *((*cpp)++);
			if (isoctal(c))
				n = (n<<3)+(c-'0');
		}
		c = n;
	}
	return(c);
}

show(cp, n)
char *cp;
{
	register int i;
	char c;

	for(i = 0; i < n; i++) {
		c = cp[i]&0177;
		if(isprint(c))
			fputc(c, stderr);
		else
			fprintf(stderr, "\\%03o", c);
	}
}

sigfunc_t waittimeout()
{
	longjmp(jmpbuf, 1);
	return;
}

static
int dogoto(label,sLine,lineCount)
char *label;
ScriptLineType *sLine;
int lineCount;
{			
	char buf[512];
	char fmt[80], cmd[32], arg1[128];
	int n;

	if(vflag)
		fprintf(stderr, "goto: %s\n", label);
	for(n = 0; n < lineCount; n++)
	{	if(equal(sLine[n].argv[0], "label")
			&& (sLine[n].argc == 2)
			&& equal(label, sLine[n].argv[1]))
			return n;
	}
	return ERR;
}

